/*
 *				**** WRP Hash ****
 *
 * This file implements a hash table
 * It uses the LinkList package
 *
 * (c) 2001 Bill Pringle, all rights reserved
 * Permission granted to use this and related files
 * for non-commercial purposes
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "linklist.h"
#include "prodlist.h"
#include "wrphash.h"

typedef struct wrphashent {
	char *key;
   void *data;
} wrpHashEnt;

static unsigned int wrpHashKey(wrpHashTable *ht, char *key);
static wrpHashEnt *wrpHashEntCreate(char *key, void *dat);
void wrpHashEntFree(wrpHashEnt *ent, void (freefcn(void *data)));
static void wrpHashAdd(wrpHashTable *ht, int val, wrpHashEnt *ent);
static wrpHashEnt *wrpHashFind(wrpHashTable *ht, char *key);
static int wrpHashFindLink(Link *link, void *key);

/**********************************************************
 * Public Routines
 *********************************************************/

/*
 * Create a new Hash Table
 */
wrpHashTable *wrpHashCreate(int maxent)
{
	wrpHashTable *ret;		/* return value */
   int i;

   ret = (wrpHashTable *)malloc(sizeof (wrpHashTable));
   if (ret != NULL)
   { /* memory allocated - initialize */
   	if (maxent < 1) maxent = WRPHASH_MAX;
      ret->maxent = maxent;	/* max number of entries */
      ret->list = create_ptr_prodlist(maxent);
      if (ret->list == NULL)
      { /* problems - out of memory */
      	free(ret);
         ret = NULL;
      }
      else
      {
      	for (i=0; i < maxent; i++)
         {
         	ret->list->item[i] = create_link_list();
         }
      }
   }

   return ret; /* return structure */
}

void wrpHashTableFree(wrpHashTable *ht, void (freefcn(void *data)))
{
	LinkList *list;
   Link *link;
   wrpHashEnt *ent;
   int i;

   if (ht == NULL) return;	/* bad args */

   for (i=0; i < ht->maxent; i++)
   { /* delete each linked list */
   	list = (LinkList *)ht->list->item[i];	/* get linked list */
      while ((link = get_head_link(list)) != NULL)
      { /* keep going until list is empty */
      	ent = delete_link(list, link);	/* get hash entry */
			wrpHashEntFree(ent, freefcn);	/* free entry */
      }
      free(list);	/* now free actual list */
      ht->list->item[i] = NULL;	/* just to be neat */
   }

   /* ProdPtrList empty, free it */
   free_ptr_prodlist(ht->list);

   /* hash table empty, now free it */
   free(ht);
}


/*
 * Add data to Hash Table
 */
void wrpHashPut(wrpHashTable *ht, char *key, void *dat)
{ /* add entry to hash table */
	unsigned int val;		/* hash value */
   wrpHashEnt *ent;

   val = wrpHashKey(ht, key);	/* get hash value */
   ent = wrpHashEntCreate(key, dat); /* create entry */
   wrpHashAdd(ht, val, ent);	/* add to hash table */
}

/*
 * Get data from Hash Table
 */
void *wrpHashGet(wrpHashTable *ht, char *key)
{
   wrpHashEnt *ret; /* we don't need this, but it makes debugging easier */

	ret = wrpHashFind(ht, key);	/* find entry matching key value */
   if (ret == NULL) return NULL;
	return (void *)ret->data;		/* return to caller */
}

/*******************************************************
 * Private Routines
 *******************************************************/

/*
 * wrpHashKey
 *
 * This function will compute a hash value from a key
 * It can be done any way you want, this is just a sample
 */
static unsigned int wrpHashKey(wrpHashTable *ht, char *key)
{ /* generate hash value from key */
	unsigned int ret;		/* has value */
   int t1=0, t2, i, maxent;

   if (ht == NULL || key == NULL) return -1; /* bad args */
   maxent = ht->maxent;
   if (maxent < 1) return -1;	/* corrupt hash table */

   for (i=0; key[i] != '\0'; i++)
   {
   	t2 = key[i];	/* get next character */
      t1 ^= t2;		/* add to lower byte */
      /* if another byte left, add to upper byte */
      if (key[i+1] != '\0')
      {
	      t2 = key[++i];	/* get next byte */
	      t2 <<= 4;			/* shift into second byte */
         t1 += t2;		/* add to upper byte */
      }
   }
   ret = t1 % ht->maxent;	/* hash value */

   return ret;	/* final hash value */
}

/*
 * wrpHashEntCreate
 *
 * This routine will create a new entry for the hash table
 */
static wrpHashEnt *wrpHashEntCreate(char *key, void *dat)
{
	wrpHashEnt *ret;		/* return value */

   ret = (wrpHashEnt *)malloc(sizeof(wrpHashEnt));
   if (ret != NULL)
   { /* initialize memory */
   	ret->key = strdup(key);
      ret->data = dat;
      if (ret->key == NULL)
      { /* problems - out of memory */
      	free(ret);
         ret = NULL;
      }
   }

   return ret;
}

void wrpHashEntFree(wrpHashEnt *ent, void (freefcn(void *data)))
{ /* free hash table entry */
	if (ent == NULL) return;	/* bad arg */
   if (ent->key != NULL) free(ent->key);	/* free key */
   freefcn(ent->data);
}


/*
 * wrpHashAdd
 *
 * Add entry to Hash Table
 */
static void wrpHashAdd(wrpHashTable *ht, int val, wrpHashEnt *ent)
{
	LinkList *list;
   Link *link;

   /* validate parameters */
   if (ht == NULL || val < 1 || ent == NULL) return;

   /* make sure there is a linked list for entry */
   if (ht->list->item[val] == NULL) return; /* should never happen */

	list = (LinkList *)ht->list->item[val];	/* get linked list */
	link = create_link_entry(ent);	/* create link for entry */
   /*
    * We are about to add this entry to the linked list
    *
    * We could add the entry anywhere
    *
    * We add to the head on the guess that more recent adds
    * will be likely to be retrieved than older ones
    *
    * We could also insert the entry into the correct place
    * within the list to keep the keys in sorted order
    */
   add_link_to_head(list, link);	/* add to list */
}

static wrpHashEnt *wrpHashFind(wrpHashTable *ht, char *key)
{
	int val;
   wrpHashEnt *ret=NULL;
   LinkList *list;
   Link *link;

   val = wrpHashKey(ht, key);	/* get hash value for key */
   list = ht->list->item[val];	/* get linked list for hash value */
	link = find_link_entry(list, key, wrpHashFindLink);
   if (link != NULL)
   { /* entry found - get it */
   	ret = (wrpHashEnt *)get_data_from_link(link);
   }

   return ret;	/* return entry */
}

static int wrpHashFindLink(Link *link, void *key)
{ /* find entry matching key - use strcmp type compare */
	wrpHashEnt *ent;

   ent = (wrpHashEnt *)get_data_from_link(link);
   return strcmp(ent->key, (char *)key);
}


